/*
 * Translate the address pointers in a BASIC program.
 * When a BASIC program is stored on a Dragon DOS disk
 * then it is tokenized and there are 9 metadata bytes
 * at the start.
 * Before each line the tokenized basic has a memory pointer to
 * the next line. Most often the program is loaded in starting at $2401,
 * but sometimes at other addresses. This program adjusts these pointers.
 *
 * Typical usage
 * ./ddosfixptr -l 0x3C01 ARTIST.BAS >newartist.bas
 */

#include <getopt.h>
#include <unistd.h>
#include <stdlib.h>
#include "ddos.h"

void usage(char *prgname) {
    fprintf(stderr, "Usage: [-p] %s <tokenized basic file>\n", prgname);
    exit(1);
}

int main(int argc, char *argv[]) {

    int state, opt, prettify = 0;
    int spaceadd = 0;
    FILE *fp;
    struct fileinfo header;
    int c,d;
    int curraddr, loadaddr;
    long int newaddr;

    if (argc < 2) {
        usage(argv[0]);
    }

    newaddr = 0;
    while ((opt = getopt(argc, argv, "l:")) != -1) {
        switch(opt) {
        case 'l':
            newaddr = strtol(optarg, NULL, 0);
            break;
        default:
            usage(argv[0]);
        }
    }

    if (optind >= argc) {
        fprintf(stderr, "Expected argument after options\n");
        exit(EXIT_FAILURE);
    }

    fp = fopen(argv[optind], "r");
    fread(&header, 9, 1, fp);
    if (header.startmark != 0x55 && header.endmark != 0xAA) {
        fprintf(stderr, "This is not a tokenized DRAGON BASIC file\n");
        exit(2);
    }
    loadaddr = header.loadaddr_h * 256 + header.loadaddr_l;
    if (newaddr == 0)
        printf("Start addr: %x\n", loadaddr);
    else {
        header.loadaddr_h = (newaddr >> 8) & 0xFF;
        header.loadaddr_l = newaddr & 0xFF;
        fwrite(&header, 9, 1, stdout);
    }
    state = 1;

    while (!feof(fp)) {
        switch (state) {
        case 1:  /* Pointer to next line */
            c = getc(fp);
            d = getc(fp);
            curraddr = c * 256 + d;
            if (curraddr == 0) state = 4;
            else state = 2;
            if (newaddr == 0)
                printf("%x\n", curraddr);
            else {
                curraddr = curraddr - loadaddr + newaddr;
                putchar((curraddr >> 8) & 0xFF);
                putchar(curraddr & 0xFF);
            }
                
            break;
        case 2:   /* Line number */
            c = getc(fp) & 0xFF;
            if (newaddr != 0) putchar(c);
            c = getc(fp) & 0xFF;
            if (newaddr != 0) putchar(c);
            state = 3;
            break;
        case 3:   /* Token or character */
            c = getc(fp) & 0xFF;
            if (newaddr != 0) putchar(c);
            if (c == 0) {
                state = 1;
                break;
            }
            if (c == 255) {
                c = getc(fp) & 0xFF;
                if (newaddr != 0) putchar(c);
            }
            break;
        default:  /* End of program */
            return 0;
        }
    }
}

